import sys
from PyQt5.QtWidgets import QWidget, QShortcut, QDesktopWidget, QApplication, QLabel, QTableWidget, QTableWidgetItem, QTableView, QLineEdit, QVBoxLayout, QHBoxLayout, QPushButton
from PyQt5.QtWidgets import QMainWindow, QAbstractItemView, QInputDialog
from PyQt5.QtCore import pyqtSlot, Qt, QAbstractTableModel, QObject
from PyQt5.QtGui import QKeySequence, QMovie, QImage, QPixmap, QFont
import glob
import re
import json
import requests
from bs4 import BeautifulSoup
import win32ui, win32gui, win32con, win32api
import os

#self.columns = ['Folder', 'Creator', 'Name', 'Code', 'Tags', 'Executable', ID]
PREFIX  = 0
FOLDER  = 1
CREATOR = 2
NAME    = 3
CODE    = 4
TAGS    = 5
EXE     = 6
NUMCOL  = 7 
#following columns aren't displayed
ID      = 7
IMG     = 8
TOTCOL  = 9

class TableModel(QAbstractTableModel):
    def __init__(self):
        QAbstractTableModel.__init__(self)
        self._data = []
        self.columns = ['Prefix', 'Folder', 'Creator', 'Name', 'Code', 'Tags', 'Executable']
        self.state = 1
        self.labels = ['Input folder regex:', 'Input creator regex  | (N)ext (P)revious', 'Input name regex  | (N)ext (P)revious', 'Input RJ code regex  | (N)ext (P)revious', 'Input tags regex  | (N)ext (P)revious', 'Executable regex  | (P)revious']
        
        count = 0
        for cur in glob.glob('.\\img\\*\\'):
            num = int(cur.split('img')[-1].strip('\\'))
            if num > count:
                count = num
        self.lastId = count
        self.curId = count+1 
        
                
    def rowCount(self, index):
            return len(self._data)
            
    def columnCount(self, index):
            #folder | creator | name | RJ | tags | executable
            return NUMCOL
            
    def data(self, index, role):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            out = self._data[index.row()][index.column()]
            if(isinstance(out,str)):
                return out
            if len(out) >= 1:
                outStr = ""
                for o in out[:-1]:
                    outStr += str(o) + ', '
                outStr += out[-1]
                return outStr
            else:
                return ""
            
    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.columns[section]
        return QAbstractTableModel.headerData(self, section, orientation, role)
        
    def flags(self, index):
        if index.column() == FOLDER:
            return Qt.ItemIsSelectable|Qt.ItemIsEnabled
        return Qt.ItemIsSelectable|Qt.ItemIsEnabled|Qt.ItemIsEditable

    def setData(self, index, value, role=Qt.EditRole):
        if role == Qt.EditRole:
            if(value == ''):
                self._data[index.row()][index.column()] = []
            #only 1 name/creator
            elif(not isinstance(value, str) and index.column() in (CREATOR,NAME,CODE)):
                self._data[index.row()][index.column()] = value[0]
            else:
                self._data[index.row()][index.column()] = value
            return True
            
    def addFolders(self, regex):
        self.curId = self.lastId +1 
        self.layoutAboutToBeChanged.emit()  
        self._data = []
        for f in glob.glob(regex.replace('[','[[]'), recursive=True):
            row = [[]] * TOTCOL
            row[FOLDER] = f
            row[ID] = self.curId
            self._data.append(row)
            self.curId += 1
        self.layoutChanged.emit()
        
    def addRegex(self, regex, indexes):
        self.layoutAboutToBeChanged.emit()  
        count = 0
        #sort indexes by row, hopefully faster than redoing the regex
        indexMap = {}
        exeRow = []
        for index in indexes:
            #don't rewrite values or set exe path with regex
            if(self._data[index.row()][index.column()] != [] or index.column() == EXE):
                continue
            r = index.row()
            if r in indexMap:
                indexMap[r] += [index]
            else:
                indexMap[r] = [index]
        for r in indexMap:
            folder = self._data[r][FOLDER]
            value = re.findall(regex, folder) #wasteful for columns with 1 value, but re.search() just doesn't work half of the time
            if(len(value) > 0):
                count += 1
                for index in indexMap[r]:
                    self.setData(index, value)
        self.layoutChanged.emit()
        
    def addStr(self, string, indexes):
        self.layoutAboutToBeChanged.emit()  
        count = 0
        #just raw indexes, I don't think optimizing per row would do anything for python
        strList = [string.strip() for string in string.split(',')]
        for index in indexes:
            if(self._data[index.row()][index.column()] != [] or index.column() == EXE):
                continue
            count += 1
            #tags and prefixes added by the str command are csv
            if(index.column() in [TAGS, PREFIX]):
                self.setData(index, strList)
            else:
                self.setData(index, string)
        self.layoutChanged.emit()

    def addExe(self, regex, indexes):
        self.layoutAboutToBeChanged.emit()  
        count = 0
        for row in [index.row() for index in indexes]:
            if(self._data[row][EXE] != []):
                continue
            folder = self._data[row][FOLDER]
            value = glob.glob((folder + regex).replace('[','[[]'), recursive=True) #glob uses brackets as escape keys
            if(len(value) > 0):
                count += 1
                for i in range(len(value)): 
                    test = value[i].lower()
                    ign = False
                    for ignore in ('config.exe', 'unitycrashhandler'):
                        if (ignore in test):
                            ign = True
                            break
                    if ign:
                        continue
                    self._data[row][EXE] = value[i]
                    break
        self.layoutChanged.emit()
        
    #WIP
    def loadData(self, filename):
        return
        
    def getIndexStr(self, row, column):
        value = self._data[row][column]
        if(isinstance(value,str)):
            return value
        if value == []:
            return None
        return value[0]
        
    def saveData(self):
        if(len(self._data) == 0):
            return
        list = {}
        for r in range(len(self._data)):
            #value to JSON
            value = {}
            value["prefix"] = self._data[r][PREFIX]
            value["folder"] = self.getIndexStr(r,FOLDER)
            value["creator"] = self.getIndexStr(r,CREATOR)
            value["name"] = self.getIndexStr(r,NAME)
            value["code"] = self.getIndexStr(r,CODE)
            value["tags"] = self._data[r][TAGS]
            value["exe"] = self._data[r][EXE]
            
            folder = './img/'+str(self._data[r][ID])+'/'
            os.mkdir(folder)
            #get icon 
            try:
                #stolen code
                ico_x = win32api.GetSystemMetrics(win32con.SM_CXICON)
                ico_y = win32api.GetSystemMetrics(win32con.SM_CYICON)

                large, small = win32gui.ExtractIconEx(self._data[r][EXE],0) 
                win32gui.DestroyIcon(large[0])

                hdc = win32ui.CreateDCFromHandle( win32gui.GetDC(0) )
                hbmp = win32ui.CreateBitmap()
                hbmp.CreateCompatibleBitmap( hdc, ico_x, ico_x )
                hdc = hdc.CreateCompatibleDC()

                hdc.SelectObject( hbmp )
                hdc.DrawIcon( (0,0), small[0] )
                hbmp.SaveBitmapFile( hdc, folder+"icon.bmp" )
            except:
                print('icon fucked up')
            #get image
            if self._data[r][IMG] != []:
                try:
                    url = self._data[r][IMG]
                    with open(folder+url.split('/')[-1], 'wb') as img:
                        img.write(requests.get(url).content)
                    value["img"] = folder+url.split('/')[-1]
                except Exception as e:
                    value["img"] = None
                    print(e)
                    print('a download fucked up')
            else:
                value["img"] = []
                
            for key in value:
                if(value[key] == None):
                    value[key] = []
            list[self._data[r][ID]] = value
        try:
            with open('g.json', 'r') as file:
                jDict = json.load(file)
                for k in jDict:
                    list[k] = jDict[k]
        except:
            pass
        file = open('g.json', 'w+')    
        try:
            file.write(json.dumps(list))
        finally:
            file.close()
        print("SAVE DONE")
            
    def rjSearch(self):
        for r in range(len(self._data)):
            #RJ code
            if self._data[r][CODE] == []:
                continue
            url = "https://www.dlsite.com/maniax/work/=/product_id/{}.html/".format(self._data[r][CODE])        
            req = requests.get(url)
            soup = BeautifulSoup(req.text, 'html.parser')
            #invalid URL
            if(soup.find('div', {'class':'error_box error_box_work'}) != None):
                continue
            #Don't overwrite
            if self._data[r][NAME] == []:
                a = soup.find('h1').text.strip()
                self._data[r][NAME] = soup.find('h1').text.strip()
            if self._data[r][CREATOR] == []:
                self._data[r][CREATOR] = soup.find('span', {'class':'maker_name'}).text.strip()
            if self._data[r][TAGS] == []:
                self._data[r][TAGS] = [soup.find('div', {'id':'category_type'}).find('a').text.strip()] + [x.text.strip() for x in soup.find('div', {'class':'main_genre'}).findAll('a')]
            #only gets the first picture, needs JS for others
            self._data[r][IMG] = "https:" + soup.find('picture').find('img')['srcset']
        print("SEARCH DONE")

#Prefix | Folder | Creator | Name | Code | Tags | Executable
DEFAULTREGEX = ['','','','','RJ\d*','','']
class regexDialog(QInputDialog):
    def __init__(self, title, message, isFolderDiag, window, dText=-1):
        super().__init__()
        #
        self.isFolderDiag = isFolderDiag        
        self.confirmed = False
        self.window = window
        #
        QShortcut(QKeySequence("Q"), self).activated.connect(self.quit)    
        #
        self.setWindowTitle(title)
        if(dText != -1):
            self.setTextValue(DEFAULTREGEX[dText])
        self.textValueChanged.connect(self.textEdited)
        self.label = QLabel(message)
        self.show() 
        
    def quit(self):
        self.close()
    
    def textEdited(self):
        self.confirmed = False
        
    def accept(self):
        if(self.isFolderDiag):
            if(not self.confirmed):
                valid, message = self.testGlob()
                if valid:
                    text = self.textValue()
                    self.setLabelText(message)
                    self.confirmed = True
                else:
                    self.setLabelText(message)   
                return
            self.window.addFolders(self.textValue())
        else:
            if(not self.confirmed):
                valid, message = self.testRegex()
                if valid:
                    text = self.textValue()
                    self.setLabelText(message)
                    self.confirmed = True
                else:
                    self.setLabelText(message)    
                return
            self.window.addRegex(self.textValue())
        self.close()

    def testGlob(self):
        try:
            folders = len(glob.glob(self.textValue().replace('[','[[]'), recursive=True))
            message = str(folders) + " folders found"
            return True, message
        except Exception as e:
            return False, str(e)            
    def testRegex(self):        
        try:
            re.compile(self.textValue())
            return True, "Valid regex"
        except Exception as e:
            return False, str(e)


class Window(QWidget):
    def __init__(self, *args, **kwargs):
        QWidget.__init__(self, *args, **kwargs)
        
        self.table = QTableView()
        self.model = TableModel()
        self.table.setModel(self.model)
        self.table.setSortingEnabled(True)
        
        self.regex = ""
        self.label = QLabel("Add (f)olders | (A)dd regex ")
        
        QShortcut(QKeySequence("r"), self).activated.connect(self.addRegexDiag)
        QShortcut(QKeySequence("s"), self).activated.connect(self.addStringDiag)
        QShortcut(QKeySequence("e"), self).activated.connect(self.addExeDiag)
        QShortcut(QKeySequence("f"), self).activated.connect(self.addFolderDiag)
        QShortcut(QKeySequence("Esc"), self).activated.connect(self.closeDiag)
                
        
        self.setWindowTitle('Editor')
        self.centralize()
        self.keepAspectRatio = True
        self.hasVideo = False
        self.videoCount1 = False
        self.setAcceptDrops(True)
        
        
        self.saveButton = QPushButton(self)
        self.saveButton.setText('Save')
        self.saveButton.clicked.connect(self.model.saveData)
                
        self.rjButton = QPushButton(self)
        self.rjButton.setText('RJ search')
        self.rjButton.clicked.connect(self.model.rjSearch)
        
        self.hBox = QHBoxLayout()
        self.hBox.addWidget(self.rjButton)
        self.hBox.addWidget(self.saveButton)
        
        self.vBox = QVBoxLayout()
        self.vBox.addWidget(self.label)
        self.vBox.addWidget(self.table)
        self.vBox.addLayout(self.hBox)
        
        self.setLayout(self.vBox)
        
        self.show()
    
    @pyqtSlot()
    def addRegexDiag(self):
        self.closeDiag()
        indexes = self.table.selectedIndexes()
        num = len(indexes)        
        if(num < 1):
            return
        c = indexes[0].column()
        for i in indexes[1:]: 
            if i.column() != c:
                c = -1
                break
        self.diag = regexDialog("Edit {} items".format(num), "Regex: | ", False, self, c)
    @pyqtSlot()
    def addFolderDiag(self):
        self.closeDiag()
        self.diag = regexDialog("Add folders", "Regex: | ", True, self)   
    @pyqtSlot()
    def addStringDiag(self):
        self.closeDiag()
        num = len(self.table.selectedIndexes())
        text, ok = QInputDialog.getText(self, "Edit {} items".format(num), 'String:', QLineEdit.Normal)
        if ok:
            self.model.addStr(text, self.table.selectedIndexes())
    @pyqtSlot()
    def addExeDiag(self):
        self.closeDiag()
        regex, ok = QInputDialog.getText(self, "Add exes", 'Regex:', QLineEdit.Normal, "*.exe")
        if ok:
            self.model.addExe('\\'+regex, self.table.selectedIndexes())
    @pyqtSlot()
    def closeDiag(self):
        try:    
            self.diag.quit()
        except:
            pass
            
    def addFolders(self, regex):
        self.model.addFolders(regex)
    def addRegex(self, regex):
        self.model.addRegex(regex, self.table.selectedIndexes())
    
    @pyqtSlot()
    def centralize(self):
        centralizedRect = self.frameGeometry()
        centralizedRect.moveCenter(QDesktopWidget().availableGeometry().center())
        self.move(centralizedRect.topLeft())
        
    @pyqtSlot()
    def exit(self):
        app.quit()


if __name__ == '__main__':
    app = QApplication([])
    window = Window()
    sys.exit(app.exec_())